/*
  CPP_CONTEST=2015
  CPP_PROBLEM=F
  CPP_LANG=CPP+OPENMP+MPI
  CPP_PROCESSES_PER_NODE=marte 1 mercurio 1
*/

/* RECORD
David Corbalán Navarro
student of Methodology of  Parallel Programming,
at the University of Murcia, Spain
February 1, 2016
time 72 msec
speed-up 520.83
*/ 

/*
 * 
 * The improvement comes basically by improving the sequential version. 
 * That's achieved by sorting the points coordinates separately and 
 * computing their cum sums in an array. Then, the distance (of any 
 * coordinate) between a point and a set of points is computed in O(n) 
 * time, where n is the size of the set to compare. Therefor, the wide 
 * of computation is the sort algorithm, which is executed in O(n log n) 
 * time for each set, where n is the size of a determinate set. Other 
 * improvements comes by suiting the data structures to take advantage 
 * of cache hierarchy usage and separating sets of data to take 
 * advantage of the cache directory protocol (assuming that protocol), 
 * because the threads don't hinder with each other.
 * 
 * The parallelism simply comes by using OpenMP in the comparison loop, 
 * scheduling dynamically this execution.
 */

#include <iostream>
#include <vector>
#include <algorithm>
#include <omp.h>
#include <mpi.h>

using namespace std;

/*
 * Conjunto de puntos.
 * 
 * Esta clase representa un conjunto de puntos ordenados en el plano.
 * Como lo que nos interesa es la distancia de Manhattan, las 
 * coordenadas X e Y se pueden ordenar independientemente.
 * De este modo, acomodamos las estructuras de datos para aprovechar 
 * la localidad espacial de la jerarquía de cachés.
 */
class SetPoints
{
	public:
		int n;
		
		vector <int> x;
		vector <int> y;
		
		vector <int> cumX;
		vector <int> cumY;
		
		SetPoints(int nPuntos, int * puntos) : n (nPuntos)
		{
			int i;
			
			for (i = 0; i < n; i++)
			{
				x.push_back(puntos[i * 2]);
				y.push_back(puntos[i * 2 + 1]);
			}
		}
		
		void compute()
		{
			sort(x.begin(), x.end());
			sort(y.begin(), y.end());
			
			int xtmp = 0;
			int ytmp = 0;
			
			for (int i = 0; i < n; i++)
			{
				xtmp += x[i];
				ytmp += y[i];
				
				cumX.push_back(xtmp);
				cumY.push_back(ytmp);
			}
		}
		
		virtual ~SetPoints()
		{
			x.clear();
			y.clear();
			cumX.clear();
			cumY.clear();
		}
		
		void print()
		{
			cout << "\tX:";
			for (int i = 0; i < n; i++)
				cout << " " << x[i];
			cout << endl;
			
			cout << "\tY:";
			for (int i = 0; i < n; i++)
				cout << " " << y[i];
			cout << endl;
			
			cout << "\tCum X:";
			for (int i = 0; i < n; i++)
				cout << " " << cumX[i];
			cout << endl;
			
			cout << "\tCum Y:";
			for (int i = 0; i < n; i++)
				cout << " " << cumY[i];
			cout << endl;
		}
} ;

float distancia(SetPoints * s1, SetPoints * s2)
{
	int sum = 0;
	
	/*
	 * Comparamos con el que menos puntos tenga en su conjunto.
	 */
	if (s2->n < s1->n)
	{
		SetPoints * aux = s1;
		s1 = s2;
		s2 = aux;
	}
	
	/*
	 * Lo hacemos para las dos dimensiones en iteraciones separadas,
	 * puesto que los cómputos son independientes (X e Y).
	 */
	for (int dim = 0; dim < 2; dim++)
	{
		int * v1 = s1->x.data();
		int * v2 = s2->x.data();
		int * cum = s2->cumX.data();
		
		if (dim)
		{
			v1 = s1->y.data();
			v2 = s2->y.data();
			cum = s2->cumY.data();
		}
		
		for (int i = 0, pos = -1; i < s1->n;)
		{
			/*
			 * Buscamos los puntos que quedan por debajo.
			 * Así, la distancia es sólo la resta de 
			 * las sumas acumuladas.
			 */
			while (pos < s2->n - 1 && v2[pos + 1] < v1[i])
				pos++;
			
			int s;
			
			/*
			 * Si no está dentro del conjunto de puntos de s2,
			 * calcula la resta de la suma acumulada. Si está 
			 * dentro, hacer ajuste a la izquierda y a la 
			 * derecha.
			 */
			if (pos == -1)
				s = abs(cum[s2->n - 1] - v1[i] * s2->n);
			else
				s = (cum[s2->n - 1] - cum[pos]) - v1[i] * (s2->n - pos - 1) + v1[i] * (pos + 1) - cum[pos];
			
			/*
			 * Si hay repetidos, no volvemos a hacer cálculos, 
			 * simplemente iteramos hasta llegar a un elemento 
			 * distinto y multiplicamos por el número de 
			 * apariciones. Y ya está.
			 */
			int rep = 0;
			
			do
			{
				i++;
				rep++;
			} while (i < s1->n && v1[i - 1] == v1[i]);
			
			sum += s * rep;
		}
	}
	
	return ((float) sum) / (s1->n * s2->n);
}

float sec(int S, int * P, int * D, int node, int np)
{
	int i, j, k;
	
	MPI_Bcast(&S, 1, MPI_INT, 0, MPI_COMM_WORLD);
	
	/*
	 * Layout del reparto.
	 * En esta sección se hace un reparto equitativo más o menos.
	 *
	 * Si el número de procesos es menos de la mitad del número 
	 * de conjuntos, se hace un reparto de manera que a cada 
	 * proceso le toca el mismo número de conjuntos.
	 * 
	 * De lo contrario, se hace un reparto de tal modo, que a 
	 * cada proceso le toca más o menos hacer las mismas 
	 * comparaciones de conjuntos, esto es, se reparte el
	 * workload.
	 */
	int * deal = new int [np];
	int totalWL = (S * (S - 1)) / 2;
	int bestWL = totalWL / np;
	if (S / np < 2)
	{
		/*
		 * Así se hace un reparto rápido 
		 * pero ineficiente en la práctica.
		 */
		for (int i = 0; i < np; i++)
			deal[i] = S / np + (i < S % np);
	}
	else
	{
		/*
		 * Aquí, el reparto se hace por workload.
		 * El workload está en base a los conjuntos, y no a sus tamaños.
		 * 
		 * Se corresponde con el número de comparaciones que toca hacer 
		 * a cada proceso, esto es:
		 * 	
		 * 	workload_total = (S * (S - 1)) / 2
		 * 
		 * El workload ideal se corresponde con el workload_total partido 
		 * entre el número de procesos:
		 * 	
		 * 	workload_ideal = workload_total / Nº de procesos
		 * 
		 * La idea es equilibrar esta función, ¿cómo? Pues muy sencillo:
		 * 
		 * 	- Se parte del workload ideal
		 * 	- Vamos cogiendo conjuntos hasta alcanzar el workload ideal
		 *	- La posición donde se detuvo coincide con el conjunto 
		 * 	  hasta el que tiene que comparar
		 */
		
		int j = 0;
		for (int i = 0; i < np - 1; i++)
		{
			int ant_j = j;
			
			int wl = 0;
			// Vamos iterando ...
			while (true)
			{
				wl += S - j - 1;
				
				// ... hasta llegar al workload ideal
				if (wl >= bestWL)
				{
					if (wl == bestWL)
						j++;
					
					break;
				}
				
				j++;
			}
			
			deal[i] = j - ant_j;
		}
		
		// El último se queda con el resto
		deal[np - 1] = S - j;
	}
	
	/*
	 * Cada nodo calcula los conjuntos que le toca
	 */
	int * setsPerNode = new int [np];
	int * setsPerNodeOffset = new int [np];
	
	// El primer nodo tiene todos los conjuntos
	setsPerNode[0] = S;
	setsPerNodeOffset[0] = 0;
	
	for (i = 1; i < np; i++)
	{
		setsPerNode[i] = setsPerNode[i - 1] - deal[i - 1];
		setsPerNodeOffset[i] = setsPerNodeOffset[i - 1] + deal[i - 1];
	}
	
	#ifdef DEBUG
	if (node == 0)
	{
		cout << "Reparto:";
		for (i = 0; i < np; i++)
			cout << " " << deal[i];
		cout << endl;
		
		cout << "Conjuntos por nodo:";
		for (i = 0; i < np; i++)
			cout << " " << setsPerNode[i];
		cout << endl;
	}
	#endif
	
	// Valor de S particular para cada nodo (local)
	S = setsPerNode[node];
	
	if (node != 0)
		P = new int [setsPerNode[node]];
	
	// Se reparte el vector de tamaños de cada conjunto
	MPI_Scatterv(P, setsPerNode, setsPerNodeOffset, MPI_INT, P, S, MPI_INT, 0, MPI_COMM_WORLD);
	
	#ifdef DEBUG
	cout << "Proceso " << (node + 1) << ":";
	for (i = 0; i < S; i++)
		cout << " " << P[i];
	cout << endl;
	#endif
	
	delete [] setsPerNode;
	delete [] setsPerNodeOffset;
	
	/*
	 * 
	 */
	int * coordsPerNode;
	int * coordsPerNodeOffset;
	
	int recvSize = 0;
	
	if (node == 0)
	{
		coordsPerNode = new int [np];
		coordsPerNodeOffset = new int [np];
		
		coordsPerNodeOffset[0] = 0;
		
		for (i = 0, k = 0; i < np; i++)
		{
			coordsPerNode[i] = 0;
			for (j = 0; j < deal[i]; j++, k++)
				coordsPerNode[i] += P[k];
			coordsPerNode[i] *= 2;
			
			if (i > 0)
				coordsPerNodeOffset[i] = coordsPerNodeOffset[i - 1] + coordsPerNode[i - 1];
			
		}
		
		for (i = np - 2; i >= 0; i--)
			coordsPerNode[i] += coordsPerNode[i + 1];
		
		#ifdef DEBUG
		for (i = 0; i < np; i++)
			cout << "Proceso " << (i + 1) << ": " << coordsPerNode[i] << " coordenadas" << endl;
		#endif
	}
	else
	{
		for (i = 0; i < S; i++)
			recvSize += P[i];
		recvSize *= 2;
		
		D = new int [recvSize];
	}
	
	MPI_Scatterv(D, coordsPerNode, coordsPerNodeOffset, MPI_INT, D, recvSize, MPI_INT, 0, MPI_COMM_WORLD);
	
	if (node == 0)
	{
		delete [] coordsPerNode;
		delete [] coordsPerNodeOffset;
	}
	
	/*
	 * Fin de las comunicaciones bestiales.
	 */
	
	int cum = 0;
	vector <SetPoints *> sets(S);
	
	/*
	 * Creamos los conjuntos tranquilamente.
	 */
	for (i = 0; i < S; i++)
	{
		sets[i] = new SetPoints (P[i], &D[cum]);
		cum += P[i] * 2;
	}
	
	/*
	 * Computamos sus estructuras de datos... tranquilamente también.
	 * Esto, ya que estamos lo hacemos en paralelo.
	 */
	#pragma omp parallel for private (i) schedule (dynamic)
	for (i = 0; i < S; i++)
		sets[i]->compute();
	
	#ifdef DEBUG
	MPI_Barrier(MPI_COMM_WORLD);
	for (i = 0; i < S; i++)
	{
		cout << "Conjunto " << (i + 1) << ":" << endl;
		sets[i]->print();
	}
	#endif
	
	float opt = 1000000.0;
	float * opts;
	
	#pragma omp parallel
	{
		int omp_node = omp_get_thread_num();
		int omp_np = omp_get_num_threads();
		
		#pragma omp single
		opts = new float [omp_np];
		
		opts[omp_node] = opt;
		
		/*
		 * Las primeras iteraciones son más grandes que las siguientes,
		 * por lo que puede haber desbalanceo de carga.
		 * Así que le dejamos a OpenMP el muerto de planificarlo con 
		 * la cláusula schedule (dynamic)
		 */
		#pragma omp for private (i, j) schedule (dynamic)
		for (i = 0; i < deal[node]; i++)
			for (j = i + 1; j < S; j++)
			{
				float res = distancia(sets[i], sets[j]);
				
				if (res < opts[omp_node])
					opts[omp_node] = res;
			}
		
		#pragma omp single
		for (i = 0; i < omp_np; i++)
			if (opts[i] < opt)
				opt = opts[i];
	}
	
	delete [] opts;
	
	for (i = 0; i < S; i++)
		delete sets[i];
	sets.clear();
	
	/*
	 * Comunicamos los resultados al nodo 0 para 
	 * que decida cuál de ellos es el mejor.
	 */
	if (node == 0)
		opts = new float [np];
	
	MPI_Gather(&opt, 1, MPI_FLOAT, opts, 1, MPI_FLOAT, 0, MPI_COMM_WORLD);
	
	if (node == 0)
	{
		for (i = 0; i < np; i++)
			if (opts[i] < opt)
				opt = opts[i];
		delete [] opts;
	}
	
	delete [] deal;
	
	if (node != 0)
	{
		delete [] P;
		delete [] D;
	}
	
	return opt;
}
